import os
import re
from datetime import datetime
from pathlib import Path
from typing import Iterable, List, Tuple

from dateutil.parser import parse

from neoteroi.mkdocs.contribs.domain import ContributionsReader, Contributor


def _read_lines_strip_comments(file_path: Path):
    with open(str(file_path), mode="rt", encoding="utf8") as file:
        lines = file.readlines()
    lines = (re.sub("#.+$", "", x).strip() for x in lines)
    return [line for line in lines if line]


class TXTContributionsReader(ContributionsReader):
    """
    A ContributionsReader that can read contributors information described in .txt
    files.
    """

    _contrib_rx = re.compile(
        r"(?P<name>[^\<]+)<(?P<email>[^\>]+)>\s\((?P<count>[^\>]+)\)"
    )

    _last_mod_time_rx = re.compile(
        r"^\s*Last\smodified\stime:\s(?P<value>.+)$", re.IGNORECASE | re.MULTILINE
    )

    def _parse_value(self, value: str) -> Tuple[str, str, int]:
        match = self._contrib_rx.search(value)
        if match:
            values = match.groupdict()
            name = values["name"].strip()
            email = values["email"].strip()
            count = int(values["count"])
        else:
            name, email, count = ("", "", -1)
        return name, email, count

    def _get_txt_file_path(self, file_path: Path) -> Path:
        path_without_extension = os.path.splitext(file_path)[0]
        return Path(path_without_extension + ".contribs.txt")

    def _get_contributors_from_txt_file(self, file_path: Path) -> Iterable[Contributor]:
        for line in _read_lines_strip_comments(file_path):
            name, email, count = self._parse_value(line)
            if name and email:
                yield Contributor(name, email, count)

    def get_contributors(self, file_path: Path) -> List[Contributor]:
        """
        Obtains the list of contributors from a txt file with the given path.
        The file contents should look like:

        Charlie Brown <charlie.brown@peanuts.com> (3)

        Having each line with such pattern:

        Name <email> (Contributions Count)

        and supporting comments using hashes:
        # Example comment
        """
        txt_path = self._get_txt_file_path(file_path)

        if not txt_path.exists():
            return []

        return list(self._get_contributors_from_txt_file(txt_path))

    def get_last_modified_date(self, file_path: Path) -> datetime:
        """Reads the last commit date of a file."""
        txt_path = self._get_txt_file_path(file_path)

        if not txt_path.exists():
            raise FileNotFoundError()

        match = self._last_mod_time_rx.search(txt_path.read_text("utf8"))

        if match:
            return parse(match.groups()[0])

        return datetime.min



"""
This module provides an implementation of cards.

neoteroi.cards

MIT License
Copyright (c) 2022 to present, Roberto Prevato
"""

from markdown import Extension

from neoteroi.mkdocs.markdown.images import Image
from neoteroi.mkdocs.markdown.processors import (
    EmbeddedBlockProcessor,
    SourceBlockProcessor,
)
from neoteroi.mkdocs.markdown.utils import create_instance, create_instances

from .domain import CardItem, Cards
from .html import CardsHTMLBuilder, CardsViewOptions


class BaseCardsProcessor:
    @property
    def name(self) -> str:
        return "cards"

    def norm_obj(self, obj):
        for item in obj:
            if "image" in item:
                item["image"] = Image.from_obj(item["image"])

    def build_html(self, parent, obj, props) -> None:
        """Builds the HTML for the given input object."""
        if not isinstance(obj, list):
            raise TypeError("Expected a list of items describing cards.")

        if self.root_config:
            new_props = dict(**self.root_config)
            new_props.update(props)
            props = new_props

        self.norm_obj(obj)
        builder = CardsHTMLBuilder(create_instance(CardsViewOptions, props))
        builder.build_html(parent, Cards(create_instances(CardItem, obj)))


class CardsEmbeddedProcessor(BaseCardsProcessor, EmbeddedBlockProcessor):
    """
    Block processor that can render cards using data embedded in the Markdown.
    """


class CardsSourceProcessor(BaseCardsProcessor, SourceBlockProcessor):
    """
    Block processor that can render cards using data from a source outside of the
    Markdown (e.g. file, URL, database).
    """


class CardsExtension(Extension):
    """Extension that includes cards."""

    def __init__(self, *args, **kwargs):
        self.config = {
            "priority": [12, "The priority to be configured for the extension."],
            "blank_target": [False, 'Whether to generate links with target="_blank"'],
        }
        super().__init__(*args, **kwargs)

    def extendMarkdown(self, md):
        md.registerExtension(self)
        priority = self.getConfig("priority")

        configs = self.getConfigs()
        del configs["priority"]

        md.parser.blockprocessors.register(
            CardsEmbeddedProcessor(md.parser).with_root_config(configs),
            "cards",
            priority + 0.1,
        )

        md.parser.blockprocessors.register(
            CardsSourceProcessor(md.parser).with_root_config(configs),
            "cards-from-source",
            priority,
        )



"""
This module defines a ContributionsReader that obtains contributors' list for a page
from the Git history. Note that this ContributionsReader won't work well in case of
history rewrites or files renamed without keeping contributor's history.
For this reason, it should be used together with a
"""

import re
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Iterable, List, Tuple

from dateutil.parser import ParserError
from dateutil.parser import parse as parse_date

from neoteroi.mkdocs.contribs.domain import ContributionsReader, Contributor


class GitContributionsReader(ContributionsReader):
    _name_email_rx = re.compile(r"(?P<name>[^\<]+)<(?P<email>[^\>]+)>")

    def _decode(self, value: bytes) -> str:
        try:
            return value.decode("utf8")
        except UnicodeDecodeError:
            return value.decode("ISO-8859-1")

    def _parse_name_and_email(self, name_and_email) -> Tuple[str, str]:
        match = self._name_email_rx.search(name_and_email)
        if match:
            name = match.groupdict()["name"].strip()
            email = match.groupdict()["email"].strip()
        else:
            name, email = ("", "")
        return name, email

    def parse_committers(self, output: str) -> Iterable[Contributor]:
        for line in output.splitlines():
            count, name_and_email = line.split("\t")
            name, email = self._parse_name_and_email(name_and_email)
            yield Contributor(name, email, int(count))

    def get_contributors(self, file_path: Path) -> List[Contributor]:
        """
        Obtains the list of contributors for a file with the given path,
        using the Git CLI.
        """
        in_process = subprocess.Popen(
            ["git", "log", "--pretty=short", str(file_path)], stdout=subprocess.PIPE
        )
        result = self._decode(
            subprocess.check_output(
                [
                    "git",
                    "shortlog",
                    "--summary",
                    "--numbered",
                    "--email",
                ],
                stdin=in_process.stdout,
            )
        )

        return list(self.parse_committers(result))

    def get_last_modified_date(self, file_path: Path) -> datetime:
        """Reads the last commit on a file."""
        result = self._decode(
            subprocess.check_output(
                ["git", "log", "-1", "--pretty=format:%ci", str(file_path)]
            )
        )
        try:
            return parse_date(result)
        except ParserError:
            return datetime.min



"""
This module provide a plugin that can be used to generate last modified time and
contributors list for each page, assuming that:

- the MkDocs site is built in a Git repository.
- the Git CLI can be used during the build.

"""

import logging
import os
from datetime import datetime
from fnmatch import fnmatch
from pathlib import Path
from subprocess import CalledProcessError
from typing import List

from mkdocs.config import config_options as c
from mkdocs.plugins import BasePlugin
from mkdocs.structure.files import File
from mkdocs.structure.pages import Page

from neoteroi.mkdocs.contribs.domain import ContributionsReader, Contributor
from neoteroi.mkdocs.contribs.git import GitContributionsReader
from neoteroi.mkdocs.contribs.html import ContribsViewOptions, render_contribution_stats
from neoteroi.mkdocs.contribs.txt import TXTContributionsReader

logger = logging.getLogger("MARKDOWN")


class DefaultContributionsReader(ContributionsReader):
    """
    Supports both contributors obtained from Git history and from configuration files.
    """

    def __init__(self) -> None:
        super().__init__()
        self._git_reader = GitContributionsReader()
        self._txt_reader = TXTContributionsReader()

    def get_contributors(self, file_path: Path) -> List[Contributor]:
        git_history_contributors = self._git_reader.get_contributors(file_path)
        configured_contributors = self._txt_reader.get_contributors(file_path)
        return list(
            {
                item.email: item
                for item in configured_contributors + git_history_contributors
            }.values()
        )

    def get_last_modified_date(self, file_path: Path) -> datetime:
        return self._git_reader.get_last_modified_date(file_path)


class ContribsPlugin(BasePlugin):
    _contribs_reader: ContributionsReader
    config_scheme = (
        ("contributors_label", c.Type(str, default="Contributors")),
        ("last_modified_label", c.Type(str, default="Last modified on")),
        ("time_format", c.Type(str, default="%Y-%m-%d %H:%M:%S")),
        ("contributors", c.Type(list, default=[])),
        ("show_last_modified_time", c.Type(bool, default=True)),
        ("show_contributors_title", c.Type(bool, default=False)),
        ("enabled_by_env", c.Type(str, default="")),
        ("exclude", c.Type(list, default=[])),
    )

    def __init__(self) -> None:
        super().__init__()
        self._contribs_reader = DefaultContributionsReader()

    def _merge_contributor_by_email(
        self,
        contributors: List[Contributor],
        contributor: Contributor,
        contributor_info: dict,
    ) -> bool:
        """
        Handles an optional "merge_with" property in the contributor info object
        (from configuration), returning true if the given contributor (from repo
        information) should be discarded (its commits count was merged with another
        one).
        """
        assert contributor.email == contributor_info.get(
            "email"
        ), "The contributor info object must match the contributor object."
        merge_with = contributor_info.get("merge_with")

        if merge_with:
            parent = next(
                (item for item in contributors if item.email == merge_with),
                None,
            )
            if parent:
                parent.count += contributor.count
                return True

        return False

    def _get_contributors(self, page_file: File) -> List[Contributor]:
        results = []
        contributors = self._contribs_reader.get_contributors(
            Path("docs") / page_file.src_path
        )
        info = self.config.get("contributors")

        if not info:
            return contributors

        for contributor in contributors:
            contributor_info = next(
                (item for item in info if item.get("email") == contributor.email), None
            )

            if contributor_info:
                contributor.image = contributor_info.get("image")
                contributor.key = contributor_info.get("key")

                if contributor_info.get("ignore"):
                    # ignore the contributor's information (can be useful for bots)
                    continue

                if (
                    "name" in contributor_info
                    and contributor_info["name"] != contributor.name
                ):
                    parent = next(
                        (
                            other
                            for other in contributors
                            if other.name == contributor_info["name"]
                        ),
                        None,
                    )
                    if parent:
                        parent.count += contributor.count
                        continue
                    else:
                        # use configured name
                        contributor.name = contributor_info["name"]

                # should contributor information be merged with another object?
                if self._merge_contributor_by_email(
                    contributors, contributor, contributor_info
                ):
                    # skip this item as it was merged with another one
                    continue

            results.append(contributor)

        return results

    def _get_last_commit_date(self, page_file: File) -> datetime:
        return self._contribs_reader.get_last_modified_date(
            Path("docs") / page_file.src_path
        )

    def _set_contributors(self, markdown: str, page: Page) -> str:
        page_file = page.file
        last_commit_date = self._get_last_commit_date(page_file)

        if last_commit_date.replace(tzinfo=None) == datetime.min:
            # The page was never committed, skip
            return markdown

        contributors = self._get_contributors(page_file)
        return (
            markdown
            + "\n\n"
            + render_contribution_stats(
                contributors,
                last_commit_date,
                ContribsViewOptions(
                    self.config["contributors_label"],
                    self.config["last_modified_label"],
                    self.config["show_last_modified_time"],
                    self.config["show_contributors_title"],
                    self.config["time_format"],
                ),
            )
        )

    def _is_ignored_page(self, page: Page) -> bool:
        if not self.config.get("exclude"):
            return False

        return any(
            fnmatch(page.file.src_path, ignored_pattern)
            for ignored_pattern in self.config["exclude"]
        )

    def _is_enabled_by_env(self):
        """
        Returns a value indicating if the plugin is enabled by env variable
        (default True if there is no env variable).
        If the user specified in the plugin configuration an env variable that controls
        if the plugin is enabled, read the env variable by that name and check if its
        value is (ci) {"1", "true"} to continue.
        """
        enabled_by_env = self.config.get("enabled_by_env")
        if enabled_by_env:
            env_var = os.environ.get(enabled_by_env)
            if env_var is None:
                return False
            return env_var.lower() in {"1", "true"}
        return True  # enabled since the user did not specify `enabled_by_env` setting

    def on_page_markdown(self, markdown, *args, **kwargs):
        if not self._is_enabled_by_env():
            return
        if self._is_ignored_page(kwargs["page"]):
            return markdown
        try:
            markdown = self._set_contributors(markdown, kwargs["page"])
        except (CalledProcessError, ValueError) as operation_error:
            logger.error(
                "Failed to display contributors list for page: %s",
                kwargs["page"].title,
                exc_info=operation_error,
            )
            pass
        return markdown
